#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>

#include <libnd.h>
#include <plugins/libnd_traceset.h>


static struct linktype_map {
	int      	dlt;
	char           *linktype;
} map[] = {
	{ DLT_NULL,		"DLT_NULL" }, 
	{ DLT_EN10MB,		"DLT_EN10MB" },
	{ DLT_EN3MB,		"DLT_EN3MB" }, 
	{ DLT_AX25,		"DLT_AX25" }, 
	{ DLT_PRONET,		"DLT_PRONET" },
	{ DLT_CHAOS,		"DLT_CHAOS" },
	{ DLT_IEEE802,		"DLT_TOKEN_RING" },
	{ DLT_ARCNET,		"DLT_ARCNET" },
	{ DLT_SLIP,		"DLT_SLIP" },
	{ DLT_PPP,		"DLT_PPP" },
	{ DLT_FDDI,	 	"DLT_FDDI" },
#ifdef DLT_ATM_RFC1483
	{ DLT_ATM_RFC1483, 	"DLT_ATM_RFC1483" },
#endif
#ifdef DLR_RAW
	{ DLT_RAW,		"DLT_RAW" },
#endif
#ifdef DLT_SLIP_BSDOS
	{ DLT_SLIP_BSDOS,	"DLT_SLIP_BSDOS" },
#endif
#ifdef DLT_PPP_BSDOS
	{ DLT_PPP_BSDOS,	"DLT_PPP_BSDOS" },
#endif
#ifdef DLT_C_HDLC
	{ DLT_C_HDLC,		"DLT_C_HDLC" },
#endif
#ifdef DLT_ATM_CLIP
	{ DLT_ATM_CLIP,		"DLT_ATM_CLIP" },
#endif
#ifdef DLT_PPP_SERIAL
	{ DLT_PPP_SERIAL,	"DLT_PPP_HDLC" },
#endif
#ifdef DLT_PPP_ETHER
	{ DLT_PPP_ETHER,	"DLT_PPP_ETHER" },
#endif
#ifdef DLT_IEEE802_11
	{ DLT_IEEE802_11,	"DLT_IEEE802_11" },
#endif
#ifdef DLT_LOOP
	{ DLT_LOOP,		"DLT_LOOP" },
#endif
#ifdef DLT_LINUX_SLL
	{ DLT_LINUX_SLL,	"DLT_LINUX_SLL" },
#endif
#ifdef DLT_TALK
	{ DLT_LTALK,		"DLT_LTALK" },
#endif
#ifdef DLT_ECONET
	{ DLT_ECONET,		"DLT_ECONET" },
#endif
#ifdef DLT_CISCO_IOS
	{ DLT_CISCO_IOS,	"DLT_CISCO_IOS" },
#endif
#ifdef DLT_PRISM_HEADER
	{ DLT_PRISM_HEADER,	"DLT_PRISM_HEADER" },
#endif
#ifdef DLT_AIRONET_HEADER
	{ DLT_AIRONET_HEADER,	"DLT_AIRONET_HEADER" },
#endif
	{ -1,			NULL }
};


const char *
name(void)
{
  return "Trace-Info";
}

const char *
author(void)
{
  return "Christian Kreibich <christian.kreibich@cl.cam.ac.uk>";
}

const char *
version(void)
{
  return "0.1";
}


gboolean
init(void)
{
  if (! libnd_plugin_find("Trace-Set"))
    {
      D(("Trace-Set plugin not found\n"));
      return FALSE;
    }
  
  return TRUE;
}


static void
ti_help(void)
{
  printf("This plugin prints general information about one or more traces.\n"
	 "\n"
	 "USAGE: lndtool -r trace-info [--names-file|-f <FILE>] [--time|-t <SEC.USEC>] [<TRACE2> <TRACE3> ...]\n"
	 "\n"
	 "  --help, -h, -?           This message.\n"
	 "  --names-file|-f FILE     Take filenames of traces from flatfile FILE.\n"
	 "  --time|-t <SEC.USEC>     Prints filenames covering given timestamp.\n");
	 
}

static gboolean
ti_trace_print(LND_TraceSet *set, LND_Trace *trace, void *unused)
{
  struct bpf_timeval delta;
  char       s[MAXPATHLEN];
  char       s2[MAXPATHLEN];
  char       s3[MAXPATHLEN];
  char      *sptr, *sptr2;
  int        i;

  libnd_tcpdump_open(trace);
  
  printf("Filename:      %s\n", trace->filename);

  g_snprintf(s, MAXPATHLEN, "%lu", (long unsigned)
	     (trace->tpm->size + sizeof (struct pcap_file_header)));

  /* Separate every three digits by a ",". First
   * reverse the string, inserting a comma every
   * three digits... */
  for (i = strlen(s) - 1, sptr = s2; i >= 0; i -= 3)
    {
      *sptr++ = s[i];
      
      if (i-1 >= 0) *sptr++ = s[i-1];	
      if (i-2 >= 0) *sptr++ = s[i-2];
      if (i-2 > 0)  *sptr++ = ',';	
    }

  /* ... and then simply reverse that string.
   */
  sptr--;
  sptr2 = s;

  do *sptr2++ = *sptr--;
  while (sptr >= s2);

  *sptr2 = '\0';

  printf("Size:          %s bytes\n", s);
  printf("Pcap version:  %i.%i\n",
	 trace->tcpdump.pfh.version_major,
	 trace->tcpdump.pfh.version_minor);
  printf("File magic:    %8X\n", trace->tcpdump.pfh.magic);
  
  for (i = 0; map[i].dlt >= 0; i++)
    {
      if (map[i].dlt == (int) trace->tcpdump.pfh.linktype)
	break;
    }
  
  if (map[i].dlt >= 0)
    printf("Linktype:      %s\n", map[i].linktype);
  else
    printf("Linktype:      unknown (%u)\n", trace->tcpdump.pfh.linktype);
  
  printf("Snaplen:       %u bytes\n", trace->tcpdump.pfh.snaplen);

  pcapnav_timeval_sub(&trace->tpm->base->end_ts,
		      &trace->tpm->base->start_ts,
		      &delta);

  libnd_misc_ctime(&trace->tpm->base->start_ts, s, MAXPATHLEN, TRUE, FALSE);
  s[strlen(s)-1] = '\0';

  libnd_misc_ctime(&trace->tpm->base->end_ts, s2, MAXPATHLEN, TRUE, FALSE);
  s2[strlen(s2)-1] = '\0';

  g_snprintf(s3, MAXPATHLEN, "%s", libnd_misc_timeval_to_string(&delta));

  printf("Start:         %10lu.%-6lu   %s\n",
	 (long unsigned) trace->tpm->base->start_ts.tv_sec,
	 (long unsigned) trace->tpm->base->start_ts.tv_usec,
	 s);

  printf("End:           %10lu.%-6lu   %s\n",
	 (long unsigned) trace->tpm->base->end_ts.tv_sec,
	 (long unsigned) trace->tpm->base->end_ts.tv_usec,
	 s2);

  printf("Span:          %10lu.%-6lu   %s\n",
	 (long unsigned) delta.tv_sec,
	 (long unsigned) delta.tv_usec,
	 s3);  

  printf("\n");

  libnd_tcpdump_close(trace);
  
  return TRUE;
  TOUCH(set);
  TOUCH(unused);
}


static gboolean
ti_trace_find(LND_TraceSet *set, LND_Trace *trace, struct bpf_timeval *ts)
{
  if (pcapnav_timeval_cmp(&trace->tpm->base->start_ts, ts) <= 0 &&
      pcapnav_timeval_cmp(ts, &trace->tpm->base->end_ts) <= 0)
    printf("%s\n", trace->filename);
  
  return TRUE;
  TOUCH(set);
}



gboolean
run(LND_Trace *unused, LND_PluginArgs *args)
{
  LND_TraceSet *traces;
  int i, files = 0;
  struct bpf_timeval ts;
  gboolean find_ts = FALSE;

  if (! (traces = libnd_traceset_new()))
    return FALSE;

  for (i = 0; i < args->argc; i++)
    {
      if (!strcmp(args->argv[i], "-h")     ||
	  !strcmp(args->argv[i], "--help") ||
	  !strcmp(args->argv[i], "-?"))
	{
	  ti_help();
	  return FALSE;
	}
      else if (!strcmp(args->argv[i], "-f") ||
	       !strcmp(args->argv[i], "--names-file"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass the name of a file that contains\n"
		     "a list of trace files, one filename per line, when\n"
		     "using the --names-file|-f option.\n");
	      return FALSE;
	    }
	  
	  if (! libnd_misc_can_read(args->argv[i]))
	    printf("Could not open '%s'\n", args->argv[i]);
	  else if (! libnd_traceset_add_trace_name_list(traces, args->argv[i]))
	    printf("Could not add all traces provided in '%s'\n", args->argv[i]);
	  else
	    files++;
	}
      else if (!strcmp(args->argv[i], "-t") ||
	       !strcmp(args->argv[i], "--time"))
	{
	  if (++i == args->argc)
	    {
	      printf("You need to pass a timestamp to look up when using\n"
		     "the --time|-t option.\n");
	      return FALSE;
	    }

	  if (sscanf(args->argv[i], "%lu.%lu", &ts.tv_sec, &ts.tv_usec) != 2)
	    {
	      printf("Please provide a timestamp in <sec>.<usec> format.\n");
	      return FALSE;
	    }

	  find_ts = TRUE;
	}
      else if (args->argv[i][0] != '-') /* Often these are just mistyped args */
	{
	  if (! libnd_traceset_add_trace_name(traces, args->argv[i]))
	    printf("Trace file %s not added.\n", args->argv[i]);
	  else
	    files++;
	}
    }
  
  if (args->argc == 0)
    {
      ti_help();
      return FALSE;
    }

  if (find_ts)    
    libnd_traceset_foreach(traces, ti_trace_find, &ts);
  else
    libnd_traceset_foreach(traces, ti_trace_print, NULL);
  
  libnd_traceset_free(traces);

  return TRUE;
  TOUCH(unused);
}
